import numpy as np


def covariance_gaussian_func(distance, range_param, sill_param):
    return sill_param * np.exp(-distance ** 2 / (2 * range_param ** 2))


def covariance_linear_func(distance, range_param, sill_param):
    if 0 <= np.abs(distance) <= range_param:
        return sill_param * (1 - (np.abs(distance)/range_param))
    return 0


def covariance_exponential_func(distance, range_param, sill_param):
    return sill_param * np.exp(-np.abs(distance)/range_param)


def covariance_spherical_func(distance, range_param, sill_param):
    if 0 <= np.abs(distance) <= range_param:
        return sill_param * (1 - (1.5 * np.abs(distance)/range_param) + (0.5 * (np.abs(distance)/range_param)**3))
    return 0



import numpy as np
import matplotlib.pyplot as plt
import matplotlib.gridspec as gridspec
from scipy.spatial.distance import squareform


class KrigingDrafter:
    pairwise_distances = None
    dissimilarity = None
    bins_x = None
    bins_y = None

    variogram_model = None
    cost_function = None
    learned_params = None
    cov_learned = None

    # class workflow flags
    _data_loaded: bool
    _variogram_created: bool
    _variogram_fitted: bool

    def summarize(self, vmin: float = None, vmax: float = None) -> None:
        """
        Plot various aspects of the kriging process: variogram, fitting
        result and covariance matrix.

        Parameters
        ----------
        vmin : float, optional
            Minimum value for color normalization in covariance matrix plot.
        vmax : float, optional
            Maximum value for color normalization in covariance matrix plot.

        Returns
        -------
        None
        """
        if not self._variogram_fitted:
            raise ValueError(f"A variogram function has not been fitted yet. Please, use the 'fit' method first")

        fig = plt.figure(figsize=(20, 6))
        gs = gridspec.GridSpec(2, 3, width_ratios=[2, 1, 1], height_ratios=[2, 2], wspace=0.1, hspace=0.1)

        ax1 = plt.subplot(gs[0, 0])
        ax2 = plt.subplot(gs[1, 0])
        ax3 = plt.subplot(gs[:, 1])

        title_size = 12
        axis_label_size = 10
        ticks_label_size = 8
        hide_tick_params = {'labelsize': 0, 'length': 0, 'colors': 'white'}

        # Variograms
        ax1.scatter(squareform(self.pairwise_distances), squareform(self.dissimilarity), color='red', marker='.', s=1,
                    label='Raw variogram')
        ax1.scatter(self.bins_x, self.bins_y, color='black', marker='x', s=100,
                    label='Binned variogram')
        ax1.set_title('Variogram', fontweight='bold', size=title_size)

        ax2.scatter(self.bins_x, self.bins_y, color='black', marker='x', s=100,
                    label='Binned variogram')
        ax2.plot(self.bins_x,
                 self.variogram_model.variogram_func(self.bins_x, *self.learned_params),
                 "--",
                 label=self.cost_function)
        ax2.set_xlabel('Distance', fontweight='bold', size=axis_label_size)

        for ax in [ax1, ax2]:
            ax.set_ylabel('Dissimilarity', fontweight='bold', size=axis_label_size)
            ax.tick_params(axis='both', which='major', labelsize=ticks_label_size)
            ax.legend(prop={'size': ticks_label_size})

        ax1.tick_params(axis='x', which='major', **hide_tick_params)

        # Covariance matrix
        im = ax3.imshow(self.cov_learned, vmin=vmin, vmax=vmax)
        ax3.set_title("Covariance Matrix", fontweight='bold', size=title_size)
        ax3.set_xlabel(f"Covariance condition number: {np.linalg.cond(self.cov_learned).round(2)}",
                       size=axis_label_size)
        ax3.tick_params(axis='both', which='major', **hide_tick_params)

        cbar = fig.colorbar(im, location='bottom', fraction=0.04)
        cbar.ax.tick_params(labelsize=ticks_label_size)

    def plot_variogram(self, show_fit: bool = False):
        """
        Plot the variogram.

        Parameters
        ----------
        show_fit : bool
            Specifies whether to plot the fitted variogram or a raw variogram.
            Default is False.

        Raises
        ------
        ValueError
            - If the 'variogram' method has not been used yet.
            - If the 'fit' method has not been used yet and the 'show_fit' parameter
            is set to True.

        Returns
        -------
        None
        """
        if not self._variogram_created:
            raise ValueError(f"A variogram has not been created yet. Please, use the 'variogram' method first")
        if show_fit and not self._variogram_fitted:
            raise ValueError(f"A variogram function has not been fitted yet. Please, use the 'fit' method first")

        fig, ax = plt.subplots(1, 1, figsize=(8, 4))
        ax.set_title('Variogram', fontweight='bold')
        ax.set_xlabel('Distance', fontweight='bold')
        ax.set_ylabel('Dissimilarity', fontweight='bold')

        if show_fit:
            ax.scatter(self.bins_x, self.bins_y, color='black',
                       marker='x', s=100, label='Binned variogram')
            x = self.bins_x.tolist()
            ax.plot(x,
                    self.variogram_model.variogram_func(x, *self.learned_params),
                    "--",
                    label=self.cost_function)
        else:
            ax.scatter(squareform(self.pairwise_distances), squareform(self.dissimilarity),
                       color='red', marker='.', s=1, label='Raw variogram')
            ax.scatter(self.bins_x, self.bins_y, color='black', marker='x', s=100,
                       label='Binned variogram')

        ax.legend()

    def plot_cov_matrix(self, vmin: float = None, vmax: float = None):
        """
        Plot the covariance matrix.

        Parameters
        ----------
        vmin : float, optional
            Minimum value for color normalization in covariance matrix plot.
        vmax : float, optional
            Maximum value for color normalization in covariance matrix plot.

        Raises
        ------
        ValueError
            - If the 'fit' method has not been used yet.

        Returns
        -------
        None
        """
        if not self._variogram_fitted:
            raise ValueError(f"Variogram function has not been fitted yet. Please, use the 'fit' method first")

        if vmin is None:
            vmin = 0

        fig, ax = plt.subplots(1, 1, figsize=(8, 4))
        im = ax.imshow(self.cov_learned, vmin=vmin, vmax=vmax)
        ax.set_title("Covariance Matrix")
        ax.set_xlabel(f"Covariance condition number: {np.linalg.cond(self.cov_learned).round(2)}")
        fig.colorbar(im)



import numpy as np

from geokrige.base import KrigingBase


class OrdinaryKriging(KrigingBase):
    def _predict_values(self):
        covariance_dist_matrix = self.variogram_model.covariance_func(self.dist_matrix, *self.learned_params)
        weights = self.cov_learned_inv @ covariance_dist_matrix

        predicted_values = self.y @ weights
        return predicted_values


class SimpleKriging(KrigingBase):
    def _predict_values(self):
        covariance_dist_matrix = self.variogram_model.covariance_func(self.dist_matrix, *self.learned_params)
        weights = self.cov_learned_inv @ covariance_dist_matrix

        ones = np.ones(self.y.shape[0])
        weights_mean = self.cov_learned_inv @ ones / self.cov_learned_inv.sum()
        mean_krig = self.y @ weights_mean

        predicted_values = mean_krig + (self.y - mean_krig) @ weights
        return predicted_values


class UniversalKriging(KrigingBase):
    def __init__(self):
        super().__init__()
        raise NotImplementedError("Universal Kriging is not supported yet.")

    def _predict_values(self):
        raise NotImplementedError

    def _create_cov_matrix(self):
        raise NotImplementedError



from typing import List

import numpy as np
from geopandas.geodataframe import GeoDataFrame
from shapely.geometry import Point, Polygon, MultiPolygon


class TransformerGDF:
    """
    A class for handling operations on a GeoDataFrame object.

    Attributes
    ----------
    X : GeoDataFrame
        An object on which the transformation will be performed.
    created_meshgrid : np.ndarray
        A mesh grid which is a result of transformation.

    Methods
    -------
    load(X)
        Load input data into the transformer.
    meshgrid(density)
        Create a meshgrid based on the loaded data.
    mask(meshgrid, gdf_object)
        Create a mask for the meshgrid based on loaded GeoDataFrame object.
    """
    def __init__(self):
        self.X = None
        self.created_meshgrid = None

    def load(self, X: GeoDataFrame) -> None:
        """
        Load input data into the transformer.

        Parameters
        ----------
        X : GeoDataFrame
            An object on which the transformation will be performed.

        Returns
        -------
        None
        """
        if not isinstance(X, GeoDataFrame):
            raise TypeError(f"Invalid input for the 'X' parameter! Should be GeoDataFrame got {X} instead")
        self.X = X

    def meshgrid(self, density: float = 1.0) -> List[np.ndarray]:
        """
        Create a mesh grid based on the loaded data.

        Parameters
        ----------
        density : float
            Grid density for 2D space. It can be increased if the returned
            mesh grid is too sparse, and decreased if it is too dense. Default
            is 1.0.

        Returns
        -------
        List[np.ndarray]
            Generated meshgrid.
        """
        if density <= 0:
            raise ValueError("The 'grid_density' value must be a positive float!")

        are_all_polygons = self.X['geometry'].apply(lambda geom: isinstance(geom, (Polygon, MultiPolygon))).all()
        if not are_all_polygons:
            raise ValueError('Not all geometry objects in the GeoDataFrame are of POLYGON type.')

        min_x, min_y, max_x, max_y = self.X.total_bounds
        x_range = max_x - min_x
        y_range = max_y - min_y

        if x_range > y_range:
            x_density_coef = density
            y_density_coef = density * (x_range / y_range)
        else:
            x_density_coef = density * (y_range / x_range)
            y_density_coef = density

        x_dots = int(round(100 * x_density_coef))
        y_dots = int(round(100 * y_density_coef))
        lon = np.linspace(min_x, max_x, num=x_dots)
        lat = np.linspace(min_y, max_y, num=y_dots)

        meshgrid = np.meshgrid(lon, lat)
        self.created_meshgrid = meshgrid
        return meshgrid

    def mask(self, meshgrid: List[np.ndarray] = None) -> np.ndarray:
        """
        Create a mask for the mesh grid based on loaded GeoDataFrame object.
        If the 'meshgrid' argument is not provided, a class attribute will
        be used (created by the 'meshgrid' method)

        Parameters
        ----------
        meshgrid : List[np.ndarray], optional
            Meshgrid to be masked. If None, uses internally stored meshgrid (created
            in the 'meshgrid' method).

        Returns
        -------
        np.ndarray
            Mask for the meshgrid. The points of the mesh grid that are within
            the GeoDataFrame polygons will have True value. The points that are
            not, will have False value.
        """
        if not meshgrid:
            meshgrid = self.created_meshgrid

        if meshgrid is None:
            raise ValueError("The 'meshgrid' parameter is None & the class 'created_meshgrid' attribute is also None. "
                             "If the 'GDFTransformer' is intended to generate a mesh grid, call the 'meshgrid' method "
                             "first. If the 'GDFTransformer' is not intended to generate a mesh grid, provide it to "
                             "the 'meshgrid' parameter.")

        geometry = [Point(lon, lat) for lon, lat in zip(*[coord.ravel() for coord in meshgrid])]
        meshgrid_as_gdf = GeoDataFrame({'geometry': geometry}, crs=str(self.X.crs))

        clipped_data = meshgrid_as_gdf.clip(self.X).copy()
        mask = np.isin(meshgrid_as_gdf.index.to_numpy(), clipped_data.index)

        return mask.reshape(meshgrid[0].shape)



from abc import abstractmethod
from typing import Union, List, Optional, Tuple
import copy

import numpy as np
from geopandas.geodataframe import GeoDataFrame
from scipy.spatial.distance import squareform
from scipy.stats import binned_statistic
from sklearn.metrics import pairwise_distances as pdist
from sklearn.preprocessing import StandardScaler

from geokrige.drafter import KrigingDrafter
from geokrige.tools.variogram_model import VariogramModel
from geokrige.utils.parameters_estimator import ParametersEstimator

from geokrige.functions.variogram import \
    variogram_gaussian_func, variogram_linear_func, \
    variogram_exponential_func, variogram_spherical_func
from geokrige.functions.covariance import \
    covariance_gaussian_func, covariance_linear_func, \
    covariance_exponential_func, covariance_spherical_func


class KrigingBase(KrigingDrafter):
    """
    Base class implementing functionality for the kriging algorithm.

    Attributes
    ----------
    VARIOGRAM_MODEL_FUNCTIONS : dict
        Dictionary containing variogram and covariance functions for different models.

    X : Union[ndarray]
        Input data loaded using the 'load' method. If a GeoDataFrame object has been
        loaded, it undergoes transformation and is then stored in this attribute as
        an ndarray object.

    y : Union[ndarray, str]
        Dependent variable values.

    scaler : StandardScaler
        StandardScaler object for data scaling.

    bins : int
        Number of bins used in variogram calculations.

    distance_metric : str
        Distance metric used for calculations.

    edges : ndarray
        Edges of bins in variogram calculations.

    pairwise_distances : ndarray
        Distance matrix between known data points.

    dissimilarity : ndarray
        Dissimilarity matrix between known data points.

    bins_x : ndarray
        Bins x-values.

    bins_y : ndarray
        Bins y-values.

    variogram_model : VariogramModel
        Variogram model.

    cost_function : str
        Name of the cost function used for variogram model fitting.

    init_args : List[int]
        Initial values for variogram model parameters.

    fixed_params : dict
        Kwargs containing fixed parameters for the variogram model.

    learned_params : List[int]
        Final values of variogram model parameters.

    cov_learned : ndarray
        Learned covariance matrix.

    cov_learned_inv : ndarray
        Inverted covariance matrix.

    dist_matrix : ndarray
        Distance matrix between data points with known value & data
        points for which a value will be estimated.

    Methods
    -------
    load(X, y)
        Load input data into the model.

    variogram(bins, distance_metric, plot)
        Compute and optionally plot the variogram for input data.

    fit(model, cost_function, init_args, plot, **fixed_params)
        Fit a variogram model to input data and optionally plot the
        fitting result.

    predict(X, distance_matrix)
        Predict values for given points.

    evaluate(groups, seed, return_)
        Evaluate model performance using cross-validation.

    summarize(vmin, vmax)
        Plot various aspects of the kriging process: variogram, fitting result
        and covariance matrix.

    plot_variogram(show_fit)
        Plot the variogram.

    plot_cov_matrix(vmin, vmax)
        Plot the covariance matrix.
    """
    VARIOGRAM_MODEL_FUNCTIONS = {
        'gaussian': (variogram_gaussian_func, covariance_gaussian_func),
        'linear': (variogram_linear_func, covariance_linear_func),
        'exp': (variogram_exponential_func, covariance_exponential_func),
        'spherical': (variogram_spherical_func, covariance_spherical_func)
    }

    def __init__(self):
        super().__init__()

        self.X: Optional[Union[np.ndarray, GeoDataFrame, list]] = None
        self.y: Optional[Union[np.ndarray, str, list]] = None
        self.scaler: Optional[StandardScaler] = None

        self.bins: Optional[int] = None
        self.distance_metric: Optional[str] = None
        self.edges: Optional[np.ndarray] = None
        self.pairwise_distances: Optional[np.ndarray] = None
        self.dissimilarity: Optional[np.ndarray] = None
        self.bins_x: Optional[np.ndarray] = None
        self.bins_y: Optional[np.ndarray] = None

        self.variogram_model: Optional[VariogramModel] = None
        self.cost_function: Optional[str] = None
        self.init_args: Optional[List[int]] = None
        self.fixed_params: Optional[dict] = None

        self.learned_params: Optional[list] = None
        self.cov_learned: Optional[np.ndarray] = None
        self.cov_learned_inv: Optional[np.ndarray] = None

        self.dist_matrix: Optional[np.ndarray] = None

        # class workflow flags
        self._data_loaded = False
        self._variogram_created = False
        self._variogram_fitted = False

    def load(self, X: Union[np.ndarray, GeoDataFrame], y: Union[np.ndarray, str]) -> None:
        """
        Load input data into the model.

        Parameters
        ----------
        X : Union[ndarray, GeoDataFrame]
            Input data. Simply put, 2D matrix when it is of type ndarray.
        y : Union[ndarray, str]
            Dependent variable values. When the 'X' parameter is of
            type GeoDataFrame, the 'y' parameter must be a string that
            specifies the column name which stores dependent variable
            in a GeoDataFrame object.

        Returns
        -------
        None
        """
        if not isinstance(X, (np.ndarray, GeoDataFrame)):
            raise TypeError("The 'X' argument must be of type 'ndarray' or 'GeoDataFrame'.")

        if isinstance(X, np.ndarray) and not isinstance(y, np.ndarray):
            raise TypeError("The 'X' argument is of type 'ndarray', so the 'y' argument must be of type 'ndarray'")

        if isinstance(X, GeoDataFrame) and not isinstance(y, str):
            raise TypeError("The 'X' argument is of type 'GeoDataFrame', so the 'y' argument must be of type 'str'.")

        if isinstance(X, GeoDataFrame):
            y = X.loc[:, y].to_numpy()
            X = self._convert_gdf_coords_to_ndarray(X)

        self.X = X
        self.y = y
        self._data_loaded = True

    def variogram(self, bins: int = 20, distance_metric: str = 'euclidean', plot: bool = True) -> None:
        """
        Compute and optionally plot the variogram for input data.

        Parameters
        ----------
        bins : int
            Number of bins used in variogram calculations. Default is 20.
        distance_metric : str
            Distance metric used for calculations. Default is 'euclidean'.
            Valid inputs: ['cityblock', 'cosine', 'euclidean', 'l1', 'l2', 'manhattan']
        plot : bool
            Specifies whether to plot the variogram. Default is True.

        Raises
        ------
        ValueError
            - If the 'load' method has not been used yet.
            - If the number of bins provided is less than or equal to 1,
            or greater than the length of the dependent variable vector.
            - If NaN values occur in the variogram y-values due to too many bins.

        Returns
        -------
        None
        """
        if not self._data_loaded:
            raise ValueError(f"Data has not been loaded to the Kriging Class. Please, use the 'load' method first")

        if bins <= 1 or bins > len(self.y):
            raise ValueError(f"Invalid input for the 'bins' parameter: {bins}. The 'bins' argument must be an integer "
                             f"greater than 1 & it must be less or equal to the length of a dependent variable vector "
                             f"-> {len(self.y)}")

        self.scaler, self.y = self._standardize_y(self.y)
        dissimilarity = (np.abs(self.y[:, None] - self.y[None, :]) ** 2) / 2
        pairwise_distances = pdist(self.X, metric=distance_metric)
        pairwise_distances = np.round(pairwise_distances, decimals=10)
        # in some cases, there might occur a computational problem when pairwise_distances is not rounded - the matrix
        # will not be symmetric

        bins_y, edges, _ = binned_statistic(squareform(pairwise_distances), squareform(dissimilarity),
                                            statistic='mean', bins=bins)
        bins_x = edges[:-1] + (edges[1] - edges[0]) / 2

        if np.isnan(bins_y).any():
            raise ValueError(f"Too many bins were created, resulting in None values occurrence in the variogram "
                             f"y-values. Specify fewer bins in the 'bins' argument. Found "
                             f"{np.isnan(bins_y).sum()} None values in the variogram y-values.")

        self.bins = bins
        self.distance_metric = distance_metric
        self.edges = edges
        self.pairwise_distances = pairwise_distances
        self.dissimilarity = dissimilarity
        self.bins_x = bins_x
        self.bins_y = bins_y

        self._variogram_created = True
        if plot:
            self.plot_variogram(show_fit=False)

    def fit(self, model: Union[VariogramModel, str] = 'gaussian', cost_function: str = 'soft_l1',
            init_args: List[int] = None, plot: bool = True, **fixed_params: float) -> None:
        """
        Fit a variogram model to input data and optionally plot the fitting result.

        Parameters
        ----------
        model : Union[VariogramModel, str]
            User's defined variogram model or a name of the model to use. Default
            is 'gaussian'. Valid inputs: [VariogramModel object, 'gaussian', 'exp',
            'spherical', 'linear']
        cost_function : str
            Cost function used during model fitting. Default is 'soft_l1'. Valid
            inputs: ['linear', 'soft_l1', 'huber', 'cauchy', 'arctan']
        init_args : List[int], optional
            Initial values for variogram model parameters. This is a starting
            point in the optimization process.
        plot : bool, optional
            Specifies whether to plot a fitting result. Default is True.
        **fixed_params : kwargs with floats, optional
            Fixed parameters for the variogram model. The keywords must have the
            same name as the variogram model parameters (built-in variogram models
            have 'distance', 'range_param', 'sill_param').

        Raises
        ------
        ValueError
            - If the 'variogram' method has not been used yet.
            - If the 'init_args' argument has more initial values than the number
            of variogram model parameters.

        Returns
        -------
        None
        """
        if not self._variogram_created:
            raise ValueError(f"A variogram has not been created yet. Please, use the 'variogram' method first")

        if model not in self.VARIOGRAM_MODEL_FUNCTIONS.keys() \
                and not isinstance(model, VariogramModel):
            raise TypeError(f"Invalid input for the 'model' parameter: expected 'VariogramModel' object "
                            f"or one of listed strings ({self.VARIOGRAM_MODEL_FUNCTIONS.keys()}), "
                            f"got {model} instead.")

        if not isinstance(init_args, (list, type(None))):
            raise TypeError(f"Invalid input for the 'init_args' parameter: should be a list or None, got "
                            f"{type(init_args)} instead")

        if isinstance(model, str):
            variogram_func, covariance_func = self.VARIOGRAM_MODEL_FUNCTIONS[model]
            model = VariogramModel()
            model.set_variogram_func(variogram_func)
            model.set_covariance_func(covariance_func)
        model = copy.deepcopy(model)

        self.variogram_model = model
        self.cost_function = cost_function
        self.init_args = init_args
        self.fixed_params = fixed_params

        if init_args is None:
            init_args = [1] * (self.variogram_model.args_numb - 1)
        if len(init_args) != (self.variogram_model.args_numb - 1):
            raise ValueError(f"Invalid input for the 'init_args' parameter: variogram model has "
                             f"{self.variogram_model.args_numb - 1} potential parameters to estimate, but "
                             f"{len(init_args)} init values were provided")

        # Parameters estimation
        for key, value in fixed_params.items():
            self.variogram_model.fix_parameter(key, value)

        weighted_values = np.histogram(squareform(self.pairwise_distances), bins=self.edges)[0].astype(float)
        estimator = ParametersEstimator(model)
        learned_params = estimator.estimate_parameters(self.bins_x, *init_args,
                                                       bins_y=self.bins_y,
                                                       weighted_values=weighted_values,
                                                       cost_function=cost_function)
        self.learned_params = learned_params

        # Covariance matrix creation
        cov_learned = self._create_cov_matrix()
        cov_learned_inv = np.linalg.pinv(cov_learned)

        self.cov_learned = cov_learned
        self.cov_learned_inv = cov_learned_inv

        self._variogram_fitted = True
        if plot:
            self.plot_variogram(show_fit=True)

    def predict(self, X: Union[List[np.ndarray], np.ndarray], distance_matrix: np.ndarray = None) -> np.ndarray:
        """
        Predict values for given points.

        Parameters
        ----------
        X : Union[List[ndarray], ndarray]
            Input data. List of matrices (numpy.meshgrid), or 2D matrix
            (numpy.ndarray).
        distance_matrix : ndarray, optional
            Distance matrix between known data points & unknown data points.
            This parameter can be used when the same distance matrix is used
            in many models (optimization feature).

        Raises
        ------
        ValueError
            - If the 'fit' method has not been used yet.
            - If the number of dimensions (coordinates) provided for prediction
            does not match the number of dimensions the model was trained on.

        Returns
        -------
        ndarray
            Predicted values.
        """
        if not self._variogram_fitted:
            raise ValueError(f"A variogram function has not been fitted yet. Please, use the 'fit' method first")

        if isinstance(X, list):  # np.meshgrid
            shape = tuple(X[0].shape)
            X = np.column_stack([matrix.ravel() for matrix in X])
        elif isinstance(X, np.ndarray):  # 2D matrix
            shape = X.shape[0]
        else:
            raise TypeError(f"The type of the 'X' argument is invalid! Valid types are [numpy.ndarray, list], "
                            f"got {type(X)} instead")

        if X.shape[1] != self.X.shape[1]:
            raise ValueError(
                f"The number of dimensions (coordinates) provided for prediction does not match the number "
                f"of dimensions the model was trained on. Training data dimensions: {self.X.shape[1]}. "
                f"Prediction data dimensions: {X.shape[1]}")

        if distance_matrix is None:
            self.dist_matrix = pdist(self.X, X, metric=self.distance_metric)
        else:
            self.dist_matrix = distance_matrix

        predicted_values = self._predict_values()
        return self.scaler.inverse_transform(predicted_values.reshape(1, -1)).reshape(shape)

    def evaluate(self, groups: int = 5, seed: int = None, return_: bool = False) -> Optional[dict]:
        """
        Evaluate model performance using cross-validation.

        Parameters
        ----------
        groups : int
            Number of groups for cross-validation. Default is 5.
        seed : int, optional
            Seed for random number generator.
        return_ : bool
            Specifies whether to return evaluation results. Default is False.

        Raises
        ------
        ValueError
            - If the 'fit' method has not been used yet.
            - If the number of 'groups' provided is less than or equal to 1,
            or greater than the number of the points on which the model was
            trained on.

        Returns
        -------
        Optional[dict]
            Evaluation results in dictionary form if return_ is True.
        """
        if not self._variogram_fitted:
            raise ValueError(f"A variogram function has not been fitted yet. Please, use the 'fit' method first")

        if groups <= 1 or groups > self.X.shape[0]:
            raise ValueError(f"Invalid input for the 'groups' parameter: {groups}. The 'groups' argument must be an "
                             f"integer greater than 1 & must not be greater than the number of the points on which "
                             f"the model was trained on -> {self.X.shape[0]}")

        if seed:
            np.random.seed(seed)

        evaluator = copy.deepcopy(self)
        data = np.column_stack([self.X, self.scaler.inverse_transform(self.y.reshape(1, -1)).ravel()])
        np.random.shuffle(data)

        mae_result = 0
        rmse_result = 0
        testing_groups = np.array_split(data, groups)
        for group in testing_groups:
            mask = np.all(np.isin(data, group), axis=1)

            train_X = data[~mask][:, 0: data.shape[1] - 1]
            train_y = data[~mask][:, -1]

            test_X = data[mask][:, 0: data.shape[1] - 1]
            test_y = data[mask][:, -1]

            # Simulating that the model was trained without testing points ---------------------------------------------
            pairwise_distances = pdist(train_X, metric=self.distance_metric)
            pairwise_distances = np.round(pairwise_distances, decimals=10)
            evaluator.pairwise_distances = pairwise_distances

            cov_learned = evaluator._create_cov_matrix()
            cov_learned_inv = np.linalg.pinv(cov_learned)
            evaluator.cov_learned = cov_learned
            evaluator.cov_learned_inv = cov_learned_inv

            dist_matrix = pdist(train_X, test_X, metric=self.distance_metric)
            evaluator.y = evaluator.scaler.transform(train_y.reshape(-1, 1)).ravel()
            # ----------------------------------------------------------------------------------------------------------

            predicted_y = evaluator.predict(test_X, dist_matrix)

            mae = np.abs(predicted_y - test_y).mean()
            rmse = np.sqrt(((predicted_y - test_y) ** 2).mean())
            mae_result += mae
            rmse_result += rmse

        if return_:
            return {'MAE': mae_result / groups, 'RMSE': rmse_result / groups}
        print(f'Mean Absolute Error: {mae_result / groups}')
        print(f'Root-Mean-Square Error: {rmse_result / groups}')

    def _create_cov_matrix(self) -> np.ndarray:
        """
        Create a covariance matrix based on learned variogram model parameters.
        Can be overridden if a kriging method requires different covariance matrix.

        Returns
        -------
        ndarray
            Created covariance matrix.
        """
        return self.variogram_model.covariance_func(self.pairwise_distances, *self.learned_params)

    @abstractmethod
    def _predict_values(self) -> np.ndarray:
        """
        Abstract method for predicting values. Must be implemented by concrete subclasses.

        Returns
        -------
        ndarray
            Predicted values.
        """
        raise NotImplementedError

    @staticmethod
    def _standardize_y(y) -> Tuple[StandardScaler, np.ndarray]:
        """
        Standardize dependent variable values.

        Parameters
        ----------
        y : ndarray
            Dependent variable values.

        Returns
        -------
        Tuple[StandardScaler, ndarray]
            Tuple containing StandardScaler object and standardized dependent variable values.
        """
        scaler = StandardScaler()
        scaler.fit(y.reshape(-1, 1))
        return scaler, scaler.transform(y.reshape(-1, 1)).ravel()

    @staticmethod
    def _convert_gdf_coords_to_ndarray(X: GeoDataFrame) -> np.ndarray:
        """
        Converts GeoDataFrame coordinates to numpy array.

        Parameters
        ----------
        X : GeoDataFrame
            Input GeoDataFrame.

        Returns
        -------
        np.ndarray
        """
        coordinates = X.get_coordinates(include_z=True)
        X = np.column_stack([coordinates[column] for column in coordinates if not coordinates[column].isna().all()])

        return X



from scipy import optimize
import numpy as np
from geokrige.tools.variogram_model import VariogramModel


class ParametersEstimator:
    """
    Estimates parameters for a variogram model based on given variogram values.

    Parameters
    ----------
    variogram_model : VariogramModel
        The variogram model for which parameters are to be estimated.

    Methods
    -------
    estimate_parameters((init_args, bins_y, weighted_values, cost_function)
        Estimates parameters for the variogram model.

    calc_residuals(*args, variogram_model, bins_y, weighted_values)
        Calculates residuals between observed and predicted variogram values.
    """
    def __init__(self, variogram_model: VariogramModel):
        self.variogram_model: VariogramModel = variogram_model

    def estimate_parameters(self, *init_args: np.ndarray, bins_y: np.ndarray = None,
                            weighted_values: np.ndarray = None, cost_function: str = None) -> np.ndarray:
        """
        Estimate parameters for the variogram model.

        Parameters
        ----------
        init_args : np.ndarray
            Initial values for the parameters to be estimated.
        bins_y : np.ndarray, optional
            Binned variogram values.
        weighted_values : np.ndarray, optional
            Weighted values for the optimization.
        cost_function : str, optional
            Name of the cost/loss function used for optimization.

        Returns
        -------
        np.ndarray
            Estimated parameters for the variogram model.

        Raises
        ------
        ValueError
            - If the input for the 'variogram_model' parameter is not an
            instance of VariogramModel.
            - If the number of initialization arguments does not match the number
            of parameters in the variogram model.
        """
        if not isinstance(self.variogram_model, VariogramModel):
            raise ValueError(
                f"Invalid input for the 'variogram_model' parameter: expected {VariogramModel} object, got "
                f"{type(self.variogram_model)} instead")

        if len([*init_args]) != self.variogram_model.args_numb:
            raise ValueError(
                f"Function takes exactly {self.variogram_model.args_numb} parameters, "
                f"but {len([*init_args])} values were provided for the init arguments")

        first_arg_name = self.variogram_model.args[0]
        self.variogram_model.fix_parameter(first_arg_name, [*init_args][0])
        params_to_estimate = [*init_args]

        indexes_to_del = []
        for key, value in self.variogram_model.fixed_params.items():
            if value is None:
                continue
            index = self.variogram_model.args.index(key)
            indexes_to_del.append(index)

        for index in sorted(indexes_to_del, reverse=True):
            del params_to_estimate[index]

        if len(params_to_estimate) == 0:
            parameters = [self.variogram_model.fixed_params[arg_name] for arg_name in self.variogram_model.args]
            return np.array(parameters[1:])  # do not return the first parameter - it differs in each step

        estimated_parameters = optimize.least_squares(
            self.calc_residuals, [*params_to_estimate],
            bounds=([0.0] * len(params_to_estimate), [np.inf] * len(params_to_estimate)),
            loss=cost_function,
            kwargs={'variogram_model': self.variogram_model, 'bins_y': bins_y,
                    'weighted_values': weighted_values}
        )

        parameters_to_return = estimated_parameters.x.tolist()
        for key, value in self.variogram_model.fixed_params.items():
            if value is None:
                continue
            index = self.variogram_model.args.index(key)
            parameters_to_return.insert(index, value)

        # do not return the first parameter (distance/lag value) that differs in each step (depending on whether we
        # are estimating parameters or calculating unknown values, e.g. in the 'predict' method)
        return parameters_to_return[1:]

    @staticmethod
    def calc_residuals(*args: np.ndarray, variogram_model: VariogramModel, bins_y: np.ndarray,
                       weighted_values: np.ndarray) -> np.ndarray:
        """
        Calculate residuals between observed and predicted variogram values.

        Parameters
        ----------
        args : tuple of np.ndarray
            Parameters for the variogram model.
        variogram_model : VariogramModel
            The variogram model.
        bins_y : np.ndarray
            Observed variogram values.
        weighted_values : np.ndarray
            Weighted values for the optimization.

        Returns
        -------
        np.ndarray
            Residuals between observed and predicted variogram values.
        """
        parameters = [*args][0].tolist()

        for key, value in variogram_model.fixed_params.items():
            if value is None:
                continue
            index = variogram_model.args.index(key)
            parameters.insert(index, value)

        residuals = variogram_model.variogram_func(*parameters) - bins_y
        return residuals * weighted_values



from inspect import isfunction
import numpy as np


class VariogramModel:
    """
    A class for defining variogram model. Variogram & covariance functions are
    defined here.

    Note that
    ---------
    Parameters in variogram and covariance functions must:
    - First parameter must be the distance parameter.
    - Be ordered consistently, i.e., if parameters for the variogram function are
      ordered as 'distance', 'sill', 'range', then parameters for the covariance
      function must be in the same order.

    Attributes
    ----------
    variogram_func : function
        The variogram function.
    covariance_func : function
        The covariance function.
    args : Tuple[str]
        Tuple containing the names of the arguments of the variogram model.
    args_numb : int
        Number of arguments of the variogram model.
    fixed_params : dict
        Dictionary containing fixed parameters of the variogram model.

    Methods
    -------
    set_variogram_func(func)
        Set the variogram function.
    set_covariance_func(func)
        Set the covariance function.
    fix_parameter(param_name, value)
        Fix a parameter of the variogram model to a specific value.
    """

    def __init__(self):
        self.variogram_func = None
        self.covariance_func = None
        self.args = None
        self.args_numb = None
        self.fixed_params = None

    def set_variogram_func(self, func) -> None:
        """
        Set the variogram function.

        Parameters
        ----------
        func : function
            The variogram function.

        Raises
        ------
        TypeError
            If the input for 'func' is not a function object.
        ValueError
            If the provided function has different arguments or argument names
            than the covariance function.

        Returns
        -------
        None
        """
        self.check_passed_function(func)

        args = func.__code__.co_varnames
        args_numb = func.__code__.co_argcount

        if self.args_numb and args_numb != self.args_numb:
            raise ValueError("Invalid input for the 'func' parameter. The function provided has different number "
                             f"of arguments than the covariance function. Variogram function has '{args_numb}'. "
                             f"Covariance function has '{self.args_numb}' arguments.")
        if self.args and self.args != args:
            raise ValueError("Invalid input for the 'func' parameter. The function provided has different argument "
                             f"names than the covariance function. Variogram function arg names: {args}. Covariance "
                             f"function arg names: {self.args}")

        self.variogram_func = np.vectorize(func)
        self.args = args
        self.args_numb = args_numb
        self.fixed_params = {}
        for arg_name in args:
            self.fixed_params[arg_name] = None

    def set_covariance_func(self, func) -> None:
        """
        Set the covariance function.

        Parameters
        ----------
        func : function
            The covariance function.

        Raises
        ------
        TypeError
            If the input for 'variogram_model' is not a function object.
        ValueError
            If the provided function has different arguments or argument names
            than the variogram function.

        Returns
        -------
        None
        """
        self.check_passed_function(func)

        args = func.__code__.co_varnames
        args_numb = func.__code__.co_argcount

        if self.args_numb and args_numb != self.args_numb:
            raise ValueError("Invalid input for the 'func' parameter. The function provided has different number "
                             f"of arguments than the variogram function. Covariance function has '{args_numb}'. "
                             f"Variogram function has '{self.args_numb}' arguments.")
        if self.args and self.args != args:
            raise ValueError("Invalid input for the 'func' parameter. The function provided has different argument "
                             f"names than the variogram function. Covariance function arg names: {args}. Variogram "
                             f"function arg names: {self.args}")

        self.covariance_func = np.vectorize(func)
        self.args = args
        self.args_numb = args_numb
        self.fixed_params = {}
        for arg_name in args:
            self.fixed_params[arg_name] = None

    def fix_parameter(self, param_name, value) -> None:
        """
        Fix a parameter of the variogram model to a specific value.

        Parameters
        ----------
        param_name : str
            The name of the parameter to fix.
        value : float
            The value to fix the parameter to (argument for the
            parameter).

        Raises
        ------
        ValueError
            If the provided parameter name is not valid.

        Returns
        -------
        None
        """
        if param_name not in self.args:
            raise ValueError(f"Invalid input for the 'param_name' parameter: '{param_name}' is not a parameter of "
                             f"the variogram model. Variogram model parameters: {self.args}")

        self.fixed_params[param_name] = value

    @staticmethod
    def check_passed_function(func) -> None:
        if not isfunction(func):
            raise TypeError(f"Invalid input for the 'variogram_model' parameter: expected a function object, got "
                            f"{type(func)} instead")



from setuptools import setup, find_packages
import codecs
import os

here = os.path.abspath(os.path.dirname(__file__))

with codecs.open(os.path.join(here, "README.md"), encoding="utf-8") as fh:
    long_description = fh.read()

setup(
    name="geokrige",
    version='0.0.0',
    author="Kamil Grala",
    author_email="grala.kamil@outlook.com",
    description="GeoKrige is a Python package designed for spatial interpolation using Kriging Methods. While "
                "primarily tailored for geospatial analysis, it is equally applicable to other spatial analysis tasks.",
    url="https://github.com/pdGruby/geokrige",
    long_description_content_type="text/markdown",
    long_description=long_description,
    packages=find_packages(),
    install_requires=[line.strip() for line in open(os.path.join(here, "requirements.txt"), "r").readlines()],
    keywords=['python', 'interpolation', 'spatial-analysis', 'kriging',
              'geospatial-analysis', 'interpolation-methods', 'geospatial-visualization'],
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",
        "Programming Language :: Python :: 3.10",
        "Operating System :: Unix",
        "Operating System :: MacOS :: MacOS X",
        "Operating System :: Microsoft :: Windows",
        "Topic :: Scientific/Engineering :: GIS"
    ],
    license='GPL-3.0'
)



import numpy as np


def variogram_gaussian_func(distance, range_param, sill_param):
    return sill_param * (1 - np.exp(-distance ** 2 / (2 * range_param ** 2)))


def variogram_linear_func(distance, range_param, sill_param):
    if 0 <= np.abs(distance) <= range_param:
        return sill_param * (np.abs(distance) / range_param)
    return sill_param


def variogram_exponential_func(distance, range_param, sill_param):
    return sill_param * (1 - np.exp(-np.abs(distance) / range_param))


def variogram_spherical_func(distance, range_param, sill_param):
    if 0 <= np.abs(distance) <= range_param:
        return sill_param * ((1.5 * np.abs(distance)/range_param) - (0.5 * (np.abs(distance)/range_param)**3))
    return sill_param



from typing import List

import rasterio
from rasterio.io import DatasetReader
import numpy as np


class TransformerRasterio:
    """
    A class for handling operations on a Rasterio object (DatasetReader).

    Attributes
    ----------
    X : DatasetReader
        An object on which the transformation will be performed.
    created_meshgrid : np.ndarray
        A mesh grid which is a result of transformation.

    Methods
    -------
    load(X)
        Load input data into the transformer.
    meshgrid(density)
        Create a mesh grid based on the loaded data.
    save(layers, path)
        Save the provided layers to a new raster file.
    """
    def __init__(self):
        self.X = None
        self.created_meshgrid = None

    def load(self, X: DatasetReader) -> None:
        """
        Load input data into the transformer.

        Parameters
        ----------
        X : DatasetReader
            An object on which the transformation will be performed.

        Returns
        -------
        None
        """
        if not isinstance(X, DatasetReader):
            raise TypeError(f"Invalid input for the 'X' parameter! Should be DatasetReader got {X} instead")
        self.X = X

    def meshgrid(self) -> List[np.ndarray]:
        """
        Create a mesh grid based on the loaded data.

        Returns
        -------
        List[np.ndarray]
            Generated meshgrid.
        """
        transform = self.X.transform
        width, height = self.X.width, self.X.height
        x = np.linspace(transform[2], transform[2] + transform[0] * (width - 1), width)
        y = np.linspace(transform[5], transform[5] + transform[4] * (height - 1), height)

        meshgrid = np.meshgrid(x, y)
        self.created_meshgrid = meshgrid

        return meshgrid

    def save(self, layers: List[np.ndarray], path: str) -> None:
        """
        Save the provided layers to a new raster file (GeoTIFF).
        The new raster file will have same attributes as the loaded
        DatasetReader object (crs, height, width etc.).

        Parameters
        ----------
        layers : List[np.ndarray]
            List of numpy arrays representing layers (mesh grids) to be
            saved.
        path : str
            A path to the raster file that will be created.

        Returns
        -------
        None
        """
        transform = self.X.transform
        width, height = self.X.width, self.X.height

        dst = rasterio.open(path, 'w', driver='GTiff', height=height, width=width, count=len(layers),
                            dtype='float32', crs=self.X.crs, transform=transform)
        i = 0
        for layer in layers:
            i += 1
            dst.write(layer, i)

        dst.close()

        print(f"Successfully saved {i} new layers into the {path} file.")
